/*
 * mod_auth_nis : Authenization with NIS (YP) system
 * Writer       : JoungKyun.Kim <http://oops.orG>
 * URL          : http://oops.org
 *
 * $Id: mod_auth_nis.c,v 1.6 2009/06/08 12:07:26 oops Exp $
 *
 * License of this module follows BSD license.
 *
 * This product includes software developed by the Apache Group
 * for use in the Apache HTTP server project (http://www.apache.org/).
 */

#define _GNU_SOURCE

#ifndef _UNISTD_H
#include <unistd.h>
#endif
#ifndef _SYS_STAT_H
#include <sys/stat.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <crypt.h>

#include "server.h"
#include "plugin.h"
#include "log.h"
#include "response.h"
#include "inet_ntop_cache.h"

#include <grp.h>
#include <pwd.h>

/*
 ******************************************************************************
 * prototype for nis
 ******************************************************************************
 */
#define YPERR_SUCCESS 0
#define YPERR_KEY 5
#define MAXSTRING 256

#define N_USER 0
#define N_PASS 1
#define N_UID  2
#define N_GID  3
#define N_NAME 4
#define N_HOME 5
#define N_SHELL 6

typedef struct {
	unsigned short debug;
	buffer * nis_domain;
	buffer * nis_usermap;
	array * nis_require;
} nis_plugin_config;

typedef struct {
	PLUGIN_DATA;

	buffer *nis_user;
	buffer *tmp_buf;

	nis_plugin_config ** config_storage;
	nis_plugin_config conf;
} nis_plugin_data;

int yp_get_default_domain (char **);
int yp_match (const char *, const char *, const char *, const int, char **, int *);
const char *yperr_string (const int);

char * http_nis_parse_entry (server *, nis_plugin_data *, char *);
static int nis_rm_ad_domain (buffer *);
static int nis_get_password (server *, nis_plugin_data *, buffer *, buffer *, buffer *);
static int nis_password_compare (server *, nis_plugin_data *, buffer *, const char *);
static int nis_match_rules (server *, nis_plugin_data *, const char *, const char *);
static int nis_user_group (server *, nis_plugin_data *, const char *, const char *);
int http_nis_check (server *, connection *, nis_plugin_data *, array *, buffer *, const char *);

/* from http_auth.c */
static unsigned char * nis_base64_decode(buffer *, const char *);

INIT_FUNC (mod_auth_nis_init) {
	nis_plugin_data *p;
	p = calloc (1, sizeof (*p));
	p->nis_user = buffer_init ();
	p->tmp_buf  = buffer_init ();

	return p;
}

FREE_FUNC (mod_auth_nis_free) {
	nis_plugin_data *p = p_d;
	UNUSED (srv);

	if ( !p ) {
		log_error_write (srv, __FILE__, __LINE__, "s",
				"FREE_FUNC: plugin_data has no data");
		return HANDLER_GO_ON;
	}

	buffer_free (p->nis_user);
	buffer_free (p->tmp_buf);

	if ( p->config_storage ) {
		size_t i;
		for ( i = 0; i < srv->config_context->used; i++ ) {
			nis_plugin_config *s = p->config_storage[i];

			if ( ! s ) continue;

			buffer_free (s->nis_domain);
			buffer_free (s->nis_usermap);
			array_free (s->nis_require);

			free (s);
		}
		free (p->config_storage);
	}

	free (p);
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC (mod_auth_nis_set_defaults) {
	nis_plugin_data *p = p_d;
	size_t i;

	config_values_t cv[] = {
		{ "nis.debug",     NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },
		{ "nis.domain",    NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ "nis.usermap",   NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ "nis.require",   NULL, T_CONFIG_LOCAL, T_CONFIG_SCOPE_CONNECTION },
		{ NULL,            NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
	};

	if ( ! p ) {
		log_error_write (srv, __FILE__, __LINE__, "s", "can't initionalize plugin_data");
		return HANDLER_ERROR;
	}

	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));

	for ( i=0; i < srv->config_context->used; i++ ) {
		nis_plugin_config *s;
		size_t n;
		data_array *da;
		array *ca;

		s = calloc (1, sizeof (nis_plugin_config));
		s->debug = 0;
		s->nis_domain  = buffer_init ();
		s->nis_usermap = buffer_init ();
		s->nis_require = array_init ();

		cv[0].destination = &(s->debug);
		cv[1].destination = s->nis_domain;
		cv[2].destination = s->nis_usermap;
		cv[3].destination = s->nis_require;

		p->config_storage[i] = s;
		ca = ((data_config *)srv->config_context->data[i])->value;

		if ( 0 != config_insert_values_global (srv, ((data_config *) srv->config_context->data[i])->value, cv) ) {
			log_error_write (srv, __FILE__, __LINE__, "s", "Can't insert global config value");
			return HANDLER_ERROR;
		}

		/*
		if ( buffer_is_empty (s->nis_domain) ) {
			log_error_write (srv, __FILE__, __LINE__, "s", "nis.domain is empty! This option must set.");
			return HANDLER_ERROR;
		}
		*/

		if ( ! s->nis_usermap->used )
			buffer_copy_string (s->nis_usermap, "passwd.byname");

		/* no nis.requrei for this section */
		if ( NULL == (da = (data_array *) array_get_element (ca, "nis.require")) ) continue;
		if ( da->type != TYPE_ARRAY ) continue;

		for ( n=0; n<da->value->used; n++ ) {
			size_t m;
			data_array *da_file = (data_array *) da->value->data[n];
			const char *realm, *require;

			if ( da->value->data[n]->type != TYPE_ARRAY ) {
				log_error_write(srv, __FILE__, __LINE__, "ss",
						"nis.require should contain an array as in:",
						"nis.require = ( \"...\" => ( ..., ...) )");
				return HANDLER_ERROR;
			}

			realm = require = NULL;

			for ( m=0; m<da_file->value->used; m++ ) {
				if ( da_file->value->data[m]->type == TYPE_STRING ) {
					if ( ! strcmp (da_file->value->data[m]->key->ptr, "realm") ) {
						realm = ((data_string *) (da_file->value->data[m]))->value->ptr;
					} else if ( ! strcmp (da_file->value->data[m]->key->ptr, "require") ) {
						require = ((data_string *) (da_file->value->data[m]))->value->ptr;
					} else {
						log_error_write (srv, __FILE__, __LINE__, "ssbs",
								"the field is unknown in:",
								"nis.require = ( \"...\" => ( ..., -> \"",
								da_file->value->data[m]->key,
								"\" <- => \"...\" ) )");

						return HANDLER_ERROR;
					}
				} else {
					log_error_write (srv, __FILE__, __LINE__, "ssbs",
							"a string was expected for:",
							"nis.require = ( \"...\" => ( ..., -> \"",
							da_file->value->data[m]->key,
							"\" <- => \"...\" ) )");

					return HANDLER_ERROR;
				}
			}

			if ( realm == NULL ) {
				log_error_write (srv, __FILE__, __LINE__, "ss",
						"the require field is missing in:",
						"nis.require = ( \"...\" => ( ..., \"realm\" => \"...\" ) )");
				return HANDLER_ERROR;
			}

			if ( require == NULL ) {
				log_error_write (srv, __FILE__, __LINE__, "ss",
						"the require field is missing in:",
						"nis.require = ( \"...\" => ( ..., \"require\" => \"...\" ) )");
				return HANDLER_ERROR;
			}

			if (realm && require) {
				data_string *ds;
				data_array *a;

				a = data_array_init();
				buffer_copy_string_buffer(a->key, da_file->key);

				ds = data_string_init();

				array_insert_unique(a->value, (data_unset *)ds);

				ds = data_string_init();

				buffer_copy_string(ds->key, "realm");
				buffer_copy_string(ds->value, realm);

				array_insert_unique(a->value, (data_unset *)ds);

				ds = data_string_init();

				buffer_copy_string(ds->key, "require");
				buffer_copy_string(ds->value, require);

				array_insert_unique(a->value, (data_unset *)ds);

				array_insert_unique(s->nis_require, (data_unset *)a);
			}
		}
	}

	return HANDLER_GO_ON;
}

#define PATCH(x) \
	p->conf.x = s->x;

static int mod_auth_nis_patch_connection (server *srv, connection *con, nis_plugin_data *p) {
	nis_plugin_config *s = p->config_storage[0];
	size_t i, j;

	PATCH (debug);
	PATCH (nis_domain);
	PATCH (nis_usermap);
	PATCH (nis_require);

	/* skip the first, the global context */
	for ( i=1; i < srv->config_context->used; i++ ) {
		data_config *dc = (data_config *) srv->config_context->data[i];
		s = p->config_storage[i];

		/* condition didn't match */
		if ( ! config_check_cond (srv, con, dc) ) continue;

		/* merge config */
		for ( j=0; j < dc->value->used; j++ ) {
			data_unset *du = dc->value->data[j];

			if ( buffer_is_equal_string (du->key, CONST_STR_LEN("nis.debug")) ) {
				PATCH (debug);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("nis.domain")) ) {
				PATCH (nis_domain);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("nis.usermap")) ) {
				PATCH (nis_usermap);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("nis.require")) ) {
				PATCH (nis_require);
			}

		}
	}

	return 0;
}
#undef PATCH

static handler_t mod_auth_nis_uri_handler (server *srv, connection *con, void *p_d) {
	size_t k;
	int nis_required = 0, nis_satisfied = 0;
	char * http_authorization = NULL;
	data_string *ds;
	nis_plugin_data *p = p_d;
	array *req;

	/* select the right config */
	mod_auth_nis_patch_connection (srv, con, p);

	if ( p->conf.nis_require == NULL ) return HANDLER_GO_ON;

	/*
	 * AUTH
	 */

	/* do we have to ask for auth ? */

	nis_required = 0;
	nis_satisfied = 0;

	/* search auth-directives for path */
	for ( k=0; k<p->conf.nis_require->used; k++ ) {
		buffer *require = p->conf.nis_require->data[k]->key;

		if ( require->used == 0 ) continue;
		if ( con->uri.path->used < require->used ) continue;

		/* if we have a case-insensitive FS we have to lower-case the URI here too */

		if ( con->conf.force_lowercase_filenames ) {
			if ( 0 == strncasecmp (con->uri.path->ptr, require->ptr, require->used - 1) ) {
				nis_required = 1;
				break;
			}
		} else {
			if ( 0 == strncmp (con->uri.path->ptr, require->ptr, require->used - 1) ) {
				nis_required = 1;
				break;
			}
		}
	}

	/* nothing to do for us */
	if ( nis_required == 0 ) return HANDLER_GO_ON;

	req = ((data_array *) (p->conf.nis_require->data[k]))->value;

	/* try to get Authorization-header */

	if ( (ds = (data_string *) array_get_element (con->request.headers, "Authorization")) != NULL )
		http_authorization = ds->value->ptr;

	if ( ds && ds->value && ds->value->used ) {
		char *nis_realm;

		/* parse auth-header */
		if ( (nis_realm = strchr (http_authorization, ' ')) != NULL )
			nis_satisfied = http_nis_check (srv, con, p, req, con->uri.path, nis_realm+1);
	}

	if ( ! nis_satisfied ) {
		data_string *realm;
		realm = (data_string *) array_get_element (req, "realm");

		con->http_status = 401;

		buffer_copy_string (p->tmp_buf, "Basic realm=\"");
		buffer_append_string_buffer (p->tmp_buf, realm->value);
		buffer_append_string (p->tmp_buf, "\"");

		response_header_insert (srv, con, CONST_STR_LEN ("WWW-Authenticate"), CONST_BUF_LEN (p->tmp_buf));

		return HANDLER_FINISHED;
	} else {
		/* the REMOTE_USER header */

		buffer_copy_string_buffer (con->authed_user, p->nis_user);
	}

	return HANDLER_GO_ON;
}

/* this function is called at dlopen() time and inits the callbacks */
int mod_auth_nis_plugin_init (plugin *p) {
	p->version          = LIGHTTPD_VERSION_ID;
	p->name             = buffer_init_string ("nis");
	p->init             = mod_auth_nis_init;
	p->set_defaults     = mod_auth_nis_set_defaults;
	p->handle_uri_clean = mod_auth_nis_uri_handler;
	p->cleanup          = mod_auth_nis_free;

	p->data             = NULL;

	return 0;
}

/* User Define API */

int http_nis_check (server *srv, connection *con, nis_plugin_data *p, array *req, buffer *url, const char *realm_str) {
	buffer * username, * password;
	char * pw;

	data_string * realm;

	realm = (data_string *) array_get_element (req, "realm");
	username = buffer_init ();

	if ( ! nis_base64_decode (username, realm_str) ) {
		log_error_write (srv, __FILE__, __LINE__, "sb", "decodeing base64-string failed", username);

		buffer_free (username);
		return 0;
	}

	/* r2 == user:password */
	if ( (pw = strchr (username->ptr, ':')) == NULL ) {
		log_error_write (srv, __FILE__, __LINE__, "sb", ": is missing in", username);

		buffer_free (username);
		return 0;
	}

	*pw++ = 0;
	username->used = pw - username->ptr;

	/*
	 * If user format is Active Directory Logon format (ex, DOMAIN\user),
	 * remove domain part.
	 */
	nis_rm_ad_domain (username);

	password = buffer_init();

	/* copy password to r1 */
	if ( nis_get_password (srv, p, username, realm->value, password) ) {
		buffer_free(username);
		buffer_free(password);

		log_error_write (srv, __FILE__, __LINE__, "s", "get_password failed");

		return 0;
	}

	/* password doesn't match */
	if ( nis_password_compare(srv, p, password, pw) ) {
		log_error_write (srv, __FILE__, __LINE__, "sbbss",
						"password doesn't match for ", con->uri.path, username,
						", IP:", inet_ntop_cache_get_ip(srv, &(con->dst_addr)));

		buffer_free(username);
		buffer_free(password);

		return 0;
	}

	/* value is our allow-rules */
	if ( nis_match_rules(srv, p, url->ptr, username->ptr) ) {
		buffer_free(username);
		buffer_free(password);

		log_error_write(srv, __FILE__, __LINE__, "s", "rules didn't match");

		return 0;
	}

	/* remember the username */
	buffer_copy_string_buffer(p->nis_user, username);

	buffer_free(username);
	buffer_free(password);

	return 1;
}

static int nis_rm_ad_domain (buffer *user) {
	char * buf;

	if ( ! user || !user->used )
		return 1;

	buf = strchr (user->ptr, '\\');
	if ( buf == NULL )
		return 0;

	buffer_copy_string (user, buf + 1);

	return 0;
}

static int nis_get_password (server *srv, nis_plugin_data *p, buffer *username, buffer *realm, buffer *password) {
	char * pwentry;
	int err, nstrlen;
	char *nstr;

	if (!username->used|| !realm->used) return -1;

	/*
	 * If nis user not specify, yp_match returns ugly return length
	 */
	if ( strlen (username->ptr) < 1 ) {
		log_error_write(srv, __FILE__, __LINE__, "s", "NIS User is not specified");
		return -1;
	}

	/* get local YP domainname */
	if ( buffer_is_empty (p->conf.nis_domain) ) {
		char nisdomain[MAXSTRING] = { 0, };

		if ( (err = yp_get_default_domain ((char **) &nisdomain)) != 0 ) {
			log_error_write(srv, __FILE__, __LINE__, "ss",
					"Can't get local YP domain:", yperr_string (err));

			return -1;
		}

		buffer_copy_string (p->conf.nis_domain, nisdomain);
	}

	if ( buffer_is_empty (p->conf.nis_usermap) )
		buffer_copy_string (p->conf.nis_usermap, "passwd.byname");

	if ( p->conf.debug ) {
		log_error_write(srv, __FILE__, __LINE__, "sb", "NIS Domain  ", p->conf.nis_domain);
		log_error_write(srv, __FILE__, __LINE__, "sb", "NIS User Map", p->conf.nis_usermap);
		log_error_write(srv, __FILE__, __LINE__, "sb", "NIS User    ", username);
	}

	err = yp_match (
			p->conf.nis_domain->ptr, p->conf.nis_usermap->ptr,
			username->ptr, strlen (username->ptr),
			&nstr, &nstrlen
	);
	if ( err == YPERR_KEY )
		err = yp_match (
				p->conf.nis_domain->ptr, p->conf.nis_usermap->ptr,
				username->ptr, strlen (username->ptr) + 1,
				&nstr, &nstrlen
		);

	if ( err == YPERR_SUCCESS )
		if ( nstr[nstrlen - 1] == 0 )
			--nstrlen;

	if ( nstrlen < 1 ) {
		log_error_write(srv, __FILE__, __LINE__, "sbs", "NIS User", username, "not found");
		return -1;
	}

	pwentry = http_nis_parse_entry (srv, p, nstr);
	buffer_copy_string(password, pwentry);

	if ( pwentry != NULL )
		free (pwentry);

	return 0;
}

char * http_nis_parse_entry (server *srv, nis_plugin_data *p, char *str) {
	char *passwd;
	char section[16];
	char **sep, **sep_t;
	int i;

	passwd = (char *) malloc (sizeof (char *) + MAXSTRING);
	sep = (char **) malloc (sizeof (char *) * 8);
	sep_t = sep;
	i = 0;
	while ( (*sep = strsep (&str, ":") ) != NULL ) {
		if ( ** sep == 0 )
			goto nextvalue;

		switch (i) {
			case N_PASS :
				memset (passwd, 0, sizeof (char *) + MAXSTRING);
				if ( strlen (*sep) >= MAXSTRING )
					strncpy (passwd, *sep, 31);
				else
					strcpy (passwd, *sep);

				if ( p->conf.debug )
					log_error_write (srv, __FILE__, __LINE__, "ss", "NIS Pass    ", passwd);

				free (sep_t);

				return passwd;
				break;
			default :
				switch (i) {
					case N_USER  : strcpy (section, "User    "); break;
					/*
					case N_UID   : strcpy (section, "Uid     "); break;
					case N_GID   : strcpy (section, "Gid     "); break;
					case N_NAME  : strcpy (section, "Name    "); break;
					case N_HOME  : strcpy (section, "Home    "); break;
					case N_SHELL : strcpy (section, "Shell   "); break;
					*/
				}
				/*
				if ( p->conf.debug )
					log_error_write(srv, __FILE__, __LINE__, "ssss", "NIS", section, *sep);
				 */
		}
		sep++;
nextvalue:
		i++;
	}
	free (sep_t);
	free (passwd);

	return NULL;
}

static int nis_password_compare (server *srv, nis_plugin_data *p, buffer *password, const char *pw) {
	UNUSED(srv);

	if ( ! pw || ! password->used )
		return 1;

	if ( p->conf.debug ) {
		log_error_write(srv, __FILE__, __LINE__, "sb", "o pw", password);
		log_error_write(srv, __FILE__, __LINE__, "ss", "n pw", crypt (pw, password->ptr));
	}

	return (strcmp (password->ptr, (char *) crypt (pw, password->ptr))) ? 1 : 0;
}

static int nis_match_rules (server *srv, nis_plugin_data *p, const char *url, const char *username) {
	const char *r = NULL, *rules = NULL;
	size_t i;
	int username_len;
	data_string *require;
	array *req;

	/* check what has to be match to fullfil the request */
	/* search auth-directives for path */
	for ( i=0; i<p->conf.nis_require->used; i++ ) {
		if ( ! p->conf.nis_require->data[i]->key->used )
			continue;

		if ( ! strncmp (url, p->conf.nis_require->data[i]->key->ptr, p->conf.nis_require->data[i]->key->used - 1))
			break;
	}

	if ( p->conf.nis_require->used == i )
		return -1;

	req = ((data_array *) (p->conf.nis_require->data[i]))->value;
	require = (data_string *) array_get_element(req, "require");

	/* if we get here, the user we got a authed user */
	if ( ! strcmp (require->value->ptr, "valid-user") )
		return 0;

	/* user=name1|group=name3|host=name4 */
	/* seperate the string by | */

	username_len = username ? strlen (username) : 0;
	r = rules = require->value->ptr;

	while ( 1 ) {
		const char *eq;
		const char *k, *v, *e;
		int k_len, v_len, r_len;
		char buf[32];
		char *pi;

		e = strchr (r, '|');
		r_len = e ? e - r : strlen (rules) - (r - rules);

		/* from r to r + r_len is a rule */

		if ( ! strncmp (r, "valid-user", r_len) ) {
			log_error_write (srv, __FILE__, __LINE__, "sb",
					"parsing the 'require' section in 'nis.require' failed: valid-user cannot be combined with other require rules",
					require->value);
			return -1;
		}

		/* search for = in the rules */
		if ( (eq = strchr(r, '=')) == NULL ) {
			log_error_write (srv, __FILE__, __LINE__, "sb",
					"parsing the 'require' section in 'nis.require' failed: a = is missing",
					require->value);
			return -1;
		}

		/* = out of range */
		if ( eq > r + r_len ) {
			log_error_write (srv, __FILE__, __LINE__, "sb",
					"parsing the 'require' section in 'nis.require' failed: = out of range",
					require->value);

			return -1;
		}

		/* the part before the = is user|group|host */
		k = r;
		k_len = eq - r;
		v = eq + 1;
		v_len = r_len - k_len - 1;

		/*
		 * cut any string after '|'
		 */
		memset (buf, 0, 32);
		strcpy (buf, v);
		pi = strchr (buf, '|');
		if ( pi )
			buf[pi - buf] = 0;
		v_len = strlen (buf);

		if ( k_len == 4 ) {
			if ( ! strncmp (k, "user", k_len) ) {
				if ( username && username_len == v_len && ! strncmp(username, buf, v_len) )
					return 0;
			} else if ( ! strncmp (k, "host", k_len) ) {
				log_error_write (srv, __FILE__, __LINE__, "s", "host ... (not implemented)");
			} else {
				log_error_write (srv, __FILE__, __LINE__, "s", "unknown key");
				return -1;
			}
		} else if ( k_len == 5 ) {
			if ( ! strncmp (k, "group", k_len) ) {
				if ( ! nis_user_group (srv, p, username, buf) )
					return 0;
			} else {
				log_error_write (srv, __FILE__, __LINE__, "ss", "unknown key", k);
				return -1;
			}
		} else {
			log_error_write (srv, __FILE__, __LINE__, "s", "unknown  key");
			return -1;
		}

		if (!e) break;
		r = e + 1;
	}

	log_error_write (srv, __FILE__, __LINE__, "s", "nothing matched");

	return -1;
}

static int nis_user_group (server *srv, nis_plugin_data *p, const char *user, const char *group) {
	struct group * entry_grp = NULL;
	struct passwd * entry_pw = NULL;
	int i = 0;

	if ( p->conf.debug ) {
		log_error_write (srv, __FILE__, __LINE__, "ss", "User  :", user);
		log_error_write (srv, __FILE__, __LINE__, "ss", "Group :", group);
	}

	/* get Group entry */
	entry_grp = getgrnam (group);

	if ( entry_grp != NULL ) {
		for ( i=0; entry_grp->gr_mem[i] != NULL; i++ ) {
			if ( ! strcmp (entry_grp->gr_mem[i], user ) ) {
				if ( p->conf.debug )
					log_error_write (srv, __FILE__, __LINE__, "sss",
									entry_grp->gr_name, "group includes", user);
				return 0;
			}
		}
	}

	/*
	 * If not found group, check group of passwd entry
	 */
	entry_pw = getpwnam (user);

	if ( entry_pw == NULL ) {
		if ( p->conf.debug )
			log_error_write (srv, __FILE__, __LINE__, "sss",
							"Can't find", user, "in the passwd entry");
		return 1;
	}

	entry_grp = getgrgid (entry_pw->pw_gid);

	if ( entry_grp == NULL ) {
		if ( p->conf.debug )
			log_error_write(srv, __FILE__, __LINE__, "sds",
							"Group who has gid", entry_pw->pw_gid, "not found");
		return 1;
	}

	if ( ! strcmp (group, entry_grp->gr_name) ) {
		if ( p->conf.debug )
			log_error_write (srv, __FILE__, __LINE__, "sds",
							"OK! Group who has gid", entry_pw->pw_gid,
							"is", entry_grp->gr_name);
		return 0;
	}

	return 1;
}

/*
 * from http_auth.c
 */
static const char nis_base64_pad = '=';

static const short nis_base64_reverse_table[256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
	-1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

static unsigned char * nis_base64_decode(buffer *out, const char *in) {
	unsigned char *result;
	int ch, j = 0, k;
	size_t i;

	size_t in_len = strlen(in);

	buffer_prepare_copy(out, in_len);

	result = (unsigned char *)out->ptr;

	ch = in[0];
	/* run through the whole string, converting as we go */
	for (i = 0; i < in_len; i++) {
		ch = in[i];

		if (ch == '\0') break;

		if (ch == nis_base64_pad) break;

		ch = nis_base64_reverse_table[ch];
		if (ch < 0) continue;

		switch(i % 4) {
			case 0:
				result[j] = ch << 2;
				break;
			case 1:
				result[j++] |= ch >> 4;
				result[j] = (ch & 0x0f) << 4;
				break;
			case 2:
				result[j++] |= ch >>2;
				result[j] = (ch & 0x03) << 6;
				break;
			case 3:
				result[j++] |= ch;
				break;
		}
	}
	k = j;
	/* mop things up if we ended on a boundary */
	if (ch == nis_base64_pad) {
		switch(i % 4) {
			case 0:
			case 1:
				return NULL;
			case 2:
				k++;
			case 3:
				result[k++] = 0;
		}
	}
	result[k] = '\0';

	out->used = k;

	return result;
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
