/*
 * mod_krisp :  COUNTRY/ISP information is set apach environment
 *              for remote addr
 * Writer    : JoungKyun.Kim <http://oops.org>
 * URL       : http://oops.org
 *
 * $Id: mod_krisp.c,v 1.5 2010/11/23 10:54:46 oops Exp $
 */

// {{{ include header files 
#include "base.h"
#include "plugin.h"
#include "log.h"
#include "response.h"

#include <krisp.h>
#include <errno.h>
// }}}

// LinuxThreads hack {{{
// If don't support NTPL, lighty is blocked pthread_create
#include <unistd.h>
#include <pthread.h>
#ifndef _CS_GNU_LIBC_VERSION
int pthread_create (pthread_t *__restrict __thread,
		__const pthread_attr_t *__restrict __attr,
		void *(*__start_routine) (void *),
		void *__restrict __arg) {
	__thread = NULL;
	__attr = NULL;
	__start_routine = NULL;
	__arg = NULL;
	return 0;
}

int pthread_join (pthread_t t, void ** tr) {
	t = 0; tr = NULL;
	return 0;
}
#endif // }}}

#define buffer_krisp(s,x) (s->x->used ? s->x->ptr : NULL)

typedef struct { // {{{ plugin_config 
	array *		proxies;
	buffer *	database;
	time_t		interval;
	bool		enabled;
	bool		debug;
} plugin_config; // }}}

typedef struct { // {{{ plugin_data 
	PLUGIN_DATA;

	KR_API *	db;

	plugin_config ** config_storage;
	plugin_config conf;
} plugin_data; // }}}

INIT_FUNC (mod_krisp_init) { // {{{
	plugin_data * p;

	p = calloc (1, sizeof (* p));

	return p;
} // }}}

FREE_FUNC (mod_krisp_free) { // {{{
	plugin_data * p = p_d;
	UNUSED (srv);

	if ( ! p ) {
		log_error_write (
			srv, __FILE__, __LINE__, "s",
			"FREE_FUNC: plugin_data has no data"
		);
		return HANDLER_GO_ON;
	}

	kr_close (&(p->db));

	if ( p->config_storage ) {
		size_t	i;
		for ( i=0; i< srv->config_context->used; i++ ) {
			plugin_config * s = p->config_storage[i];

			if ( ! s )
				continue;

			array_free (s->proxies);
			buffer_free (s->database);
			free (s);
		}
	}

	free (p->config_storage);
	return HANDLER_GO_ON;
} // }}}

SETDEFAULTS_FUNC (mod_krisp_set_defaults) { // {{{
	plugin_data *	p = p_d;
	size_t			i;
	char			err[1024] = { 0, };

	config_values_t cv[] = {
		{ "krisp.enabled",       NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },
		{ "krisp.debug",         NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },
		{ "krisp.interval",      NULL, T_CONFIG_SHORT,   T_CONFIG_SCOPE_SERVER },
		{ "krisp.database-path", NULL, T_CONFIG_STRING,  T_CONFIG_SCOPE_SERVER },
		{ "krisp.user-proxies",  NULL, T_CONFIG_ARRAY,   T_CONFIG_SCOPE_SERVER },
		{ NULL,                  NULL, T_CONFIG_UNSET,   T_CONFIG_SCOPE_UNSET }
	};

	if ( ! p ) {
		log_error_write (
			srv, __FILE__, __LINE__, "ss",
			__func__, "=> Can't initionalize plugin_data"
		);
		return HANDLER_ERROR;
	}

	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));

	p->db = NULL;
	for (i = 0; i < srv->config_context->used; i++) {
		plugin_config *	s;

		s = calloc (1, sizeof (plugin_config));
		s->proxies = array_init ();
		s->database = buffer_init ();
		s->interval = 3600;
		s->enabled = false;
		s->debug = false;

		cv[0].destination = &(s->enabled);
		cv[1].destination = &(s->debug);
		cv[2].destination = &(s->interval);
		cv[3].destination = s->database;
		cv[4].destination = s->proxies;

		p->config_storage[i] = s;

		if ( config_insert_values_global (srv, ((data_config *)srv->config_context->data[i])->value, cv) != 0 ) {
			log_error_write (
				srv, __FILE__, __LINE__, "s",
				"Can't insert global config value"
			);
			return HANDLER_ERROR;
		}

		// {{{ Debug messages:: Configuration settings
		if ( s->debug == true ) {
			size_t k;

			log_error_write (
				srv, __FILE__, __LINE__, "sd", "Set Engine On      : ", s->enabled
			);
			log_error_write (
				srv, __FILE__, __LINE__, "ss",
				"Set database       : ",
				buffer_krisp (s, database)
			);
			log_error_write (
				srv, __FILE__, __LINE__, "sd", "Set mtime interval : ", s->interval
			);

			for ( k=0; k<s->proxies->used; k++ ) {
				data_string *	ds = (data_string *) s->proxies->data[k];
				log_error_write (
					srv, __FILE__, __LINE__, "ss", "Set User Proxies   : ", ds->value->ptr
				);
			}
		} // }}}

		if ( s->enabled == false )
			return HANDLER_GO_ON;

		if ( s->debug == true ) { // {{{ Debug Messages
			log_error_write (
				srv, __FILE__, __LINE__, "s", "Connect Database ::"
			);
		} // }}}

		if ( p->db != NULL )
			continue;

		if ( kr_open (&(p->db), buffer_krisp (s, database), err) == false ) {
			log_error_write (
				srv, __FILE__, __LINE__, "ss",
				"Connect error:", err
			);
			s->enabled = false;
			return HANDLER_GO_ON;
		}

		p->db->db_time_stamp_interval = s->interval;
		if ( s->debug == true ) { // {{{ Debug Messages
			log_error_write (
				srv, __FILE__, __LINE__, "sds",
				"Set Database mtime check interval is", s->interval, "/sec"
			);
		} // }}}
	}

	return HANDLER_GO_ON;
} // }}}

#define PATCH(x) p->conf.x = s->x;
static int mod_krisp_patch_connection (server * srv, connection * con, plugin_data * p) { // {{{
	plugin_config * s = p->config_storage[0];
	size_t i, j;

	PATCH (debug);
	PATCH (enabled);
	PATCH (proxies);

	/* skip the first, the global context */
	for ( i=1; i < srv->config_context->used; i++ ) {
		data_config *dc = (data_config *) srv->config_context->data[i];
		s = p->config_storage[i];

		/* condition didn't match */
		if ( ! config_check_cond (srv, con, dc) ) continue;

		/* merge config */
		for ( j=0; j < dc->value->used; j++ ) {
			data_unset *du = dc->value->data[j];
			if ( buffer_is_equal_string (du->key, CONST_STR_LEN("krisp.user_proxies")) ) {
				PATCH (proxies);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("krisp.enabled")) ) {
				PATCH (enabled);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("krisp.debug")) ) {
				PATCH (debug);
			}
		}
	}

	return 0;
} // }}}
#undef PATCH

URIHANDLER_FUNC(mod_krisp_uri_handler) { // {{{
	plugin_data *	p = p_d;
	data_string *	px;
	data_string *	rq;
	KRNET_API		isp;
	size_t			i;

	/* select the right config */
	mod_krisp_patch_connection (srv, con, p);

	if ( p->conf.debug == true ) { // {{{ Debug Messages
		log_error_write (srv, __FILE__, __LINE__, "ss", "***", __func__);
	} // }}}

	if ( p->conf.enabled == false )
		return HANDLER_GO_ON;

	SAFECPY_256 (isp.ip, con->dst_addr_buf->ptr);
	for ( i=0; i<p->conf.proxies->used; i++ ) {
		if ( ! con->request.headers->used )
			break;

		px = (data_string *) p->conf.proxies->data[i];
		if ( p->conf.debug == true ) { // {{{ Debug Messages
			log_error_write (srv, __FILE__, __LINE__, "ss", "Search Key =>", px->value->ptr);
		} // }}}
		rq = (data_string *) array_get_element (con->request.headers, px->value->ptr);
		if ( p->conf.debug == true ) { // {{{ Debug Messages
			log_error_write (srv, __FILE__, __LINE__, "ss", "Search Value =>", rq ? rq->value->ptr : "--");
		} // }}}

		if ( rq != NULL && rq->value->used ) {
			char * comma_ptr;
			char * new_addr;

			new_addr = strdup (rq->value->ptr);
			comma_ptr = strchr (new_addr, ',');
			if ( comma_ptr != NULL )
				*comma_ptr = 0;

			if ( p->conf.debug == true ) { // {{{ Debug Messages
				log_error_write (
					srv, __FILE__, __LINE__, "ssss",
					"Set Alternative IP[", rq->key->ptr, "] :", new_addr
				);
			} // }}}
			SAFECPY_256 (isp.ip, new_addr);
			free (new_addr);
			break;
		}
	}

	isp.verbose = false;
	if ( kr_search (&isp, p->db) ) {
		log_error_write (
			srv, __FILE__, __LINE__, "ss",
			"Search error:", isp.err
		);
		return HANDLER_GO_ON;
	}

	if ( p->conf.debug == true ) { // {{{ Debug Messages
		log_error_write (srv, __FILE__, __LINE__, "ss", "Original IP   ::", con->dst_addr_buf->ptr);
		log_error_write (srv, __FILE__, __LINE__, "ss", "Search IP     ::", isp.ip);
		log_error_write (srv, __FILE__, __LINE__, "ss", "Search Start  ::", long2ip (isp.start));
		log_error_write (srv, __FILE__, __LINE__, "ss", "Search End    ::", long2ip (isp.end));
		log_error_write (srv, __FILE__, __LINE__, "ss", "Search ISP    ::", isp.icode);
		log_error_write (srv, __FILE__, __LINE__, "ss", "Search Country::", isp.ccode);
	} // }}}

	{ // {{{ Set environments
		data_string * ds;

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_ORIGINAL_IP", 17);
		buffer_copy_string_len (ds->value, con->dst_addr_buf->ptr, con->dst_addr_buf->size);
		array_insert_unique (con->environment, (data_unset *) ds);

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_CHECK_IP", 14);
		buffer_copy_string_len (ds->value, isp.ip, strlen (isp.ip));
		array_insert_unique (con->environment, (data_unset *) ds);

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_ISP_CODE", 14);
		buffer_copy_string_len (ds->value, isp.icode, strlen (isp.icode));
		array_insert_unique (con->environment, (data_unset *) ds);

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_ISP_NAME", 14);
		buffer_copy_string_len (ds->value, isp.iname, strlen (isp.iname));
		array_insert_unique (con->environment, (data_unset *) ds);

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_COUNTRY_CODE", 18);
		buffer_copy_string_len (ds->value, isp.ccode, strlen (isp.ccode));
		array_insert_unique (con->environment, (data_unset *) ds);

		if ( NULL == (ds = (data_string *) array_get_unused_element (con->environment, TYPE_STRING)) )
			ds = data_string_init ();
		buffer_copy_string_len (ds->key, "KRISP_COUNTRY_NAME", 18);
		buffer_copy_string_len (ds->value, isp.cname, strlen (isp.cname));
		array_insert_unique (con->environment, (data_unset *) ds);
	} // }}}

	if ( p->conf.debug == true ) { // {{{ Debug Messages
		data_string * ds = (data_string *) array_get_element (con->environment, "KRISP_ORIGINAL_IP");
		log_error_write (
			srv, __FILE__, __LINE__, "ss",
			"KRISP_ORIGINAL_IP::",
			ds->value->used ? ds->value->ptr : "--"
		);

		ds = (data_string *) array_get_element (con->environment, "KRISP_ISP_CODE");
		log_error_write (
			srv, __FILE__, __LINE__, "ss",
			"KRISP_ISP_CODE::",
			ds->value->used ? ds->value->ptr : "--"
		);
	} // }}}

	return HANDLER_GO_ON;
} // }}}

/* this function is called at dlopen() time and inits the callbacks */
int mod_krisp_plugin_init (plugin *p) { // {{{
	p->version         = LIGHTTPD_VERSION_ID;
	p->name            = buffer_init_string ("krisp");
	p->init            = mod_krisp_init;
	p->set_defaults    = mod_krisp_set_defaults;
	p->handle_uri_raw  = mod_krisp_uri_handler;
	p->cleanup         = mod_krisp_free;

	p->data             = NULL;

	return 0;
} // }}}

/*
 * User Define APIs
 */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */

