/*
 * mod_net_access : Network Access control as Apache
 * Writer      : JoungKyun.Kim <http://oops.org>
 * URL         : http://oops.org
 *
 * $Id: mod_net_access.c,v 1.2 2008/11/04 13:45:46 oops Exp $
 *
 * License of this module follows BSD license.
 */

#define _GNU_SOURCE

#include "base.h"
#include "log.h"
#include "buffer.h"

#include "plugin.h"

/*
 ******************************************************************************
 * prototype for net_access
 ******************************************************************************
 */
typedef struct {
	unsigned short debug;
	buffer * order;
	array * list;
} plugin_config;

typedef struct {
	PLUGIN_DATA;

	plugin_config ** config_storage;
	plugin_config conf;
} plugin_data;

/*
 ******************************************************************************
 * User Define API
 ******************************************************************************
 */

typedef struct {
	server *srv;
	buffer *dst;
	buffer *src;
	short debug;
} E_Match;

short naccess_ip_match (E_Match ip) {
	char ip_t[33] = { 0, };
	char * netmask;
	struct in_addr dst_ip;

	if ( ! ip.dst->used || ! ip.src->used )
		return 1;

	if ( ! inet_aton (ip.dst->ptr, &dst_ip) )
		return 1;

	if ( ip.debug ) {
		log_error_write (ip.srv, __FILE__, __LINE__, "s", "-- Network matches information");
		log_error_write (ip.srv, __FILE__, __LINE__, "ss", "Destination IP:", ip.dst->ptr);
		log_error_write (ip.srv, __FILE__, __LINE__, "ss", "Source      IP:", ip.src->ptr);
	}

	strcpy (ip_t, ip.src->ptr);
	netmask = strrchr (ip_t, '/');

	if ( netmask == NULL )
		return strcmp (ip.dst->ptr, ip.src->ptr) ? 1 : 0;
	else {
		struct in_addr src_ip, src_nm;
		long nm;
		int nm_bits;
		char *err;

		*netmask = 0;

		nm_bits = strtol (netmask + 1, &err, 10);
		if ( *err ) {
			if ( ! inet_aton (netmask + 1, &src_nm) )
				return 1;

			nm = ntohl (src_nm.s_addr);
		} else {
			nm = htonl (~((1 << (32 - nm_bits)) - 1));
			nm = ntohl (nm);
		}

		if ( ip.debug ) {
			src_nm.s_addr = htonl (nm);
			log_error_write (ip.srv, __FILE__, __LINE__, "ss", "Netmask       :", inet_ntoa (src_nm));
			log_error_write (ip.srv, __FILE__, __LINE__, "ss", "Parsed Src IP :", ip_t);
		}

		if ( ! inet_aton (ip_t, &src_ip) )
			return 1;

		if ( ip.debug )
			log_error_write (ip.srv, __FILE__, __LINE__, "ss", "Converted IP :", inet_ntoa (src_ip));

		if ( (ntohl (src_ip.s_addr) & nm) == (ntohl (dst_ip.s_addr) & nm) )
			return 0;
	}

	return 1;
}

short naccess_match (server *srv, connection *c, plugin_config p) {
	unsigned int i;
	E_Match ips;

	if ( ! c->dst_addr_buf->used )
		return 1;

	ips.srv   = srv;
	ips.src   = buffer_init ();
	ips.dst   = buffer_init ();
	ips.debug = p.debug;

	for ( i=0; i<p.list->used; i++ ) {
		data_string *ds = (data_string *) p.list->data[i];

		if ( ! ds->value->used )
			continue;

		ips.src   = buffer_init ();
		ips.dst   = buffer_init ();

		buffer_copy_string_buffer (ips.src, ds->value);
		buffer_copy_string_buffer (ips.dst, c->dst_addr_buf);

		if ( ! naccess_ip_match (ips) ) {
			buffer_free (ips.src);
			buffer_free (ips.dst);
			return 0;
		}

		buffer_free (ips.src);
		buffer_free (ips.dst);
	}
	
	return 1;
}

/*
 ******************************************************************************
 * Lighttpd eAccess Module Interface
 ******************************************************************************
 */

INIT_FUNC (mod_net_access_init) {
	plugin_data *p;
	p = calloc (1, sizeof (*p));

	return p;
}

FREE_FUNC (mod_net_access_free) {
	plugin_data *p = p_d;
	UNUSED (srv);

	if ( !p ) {
		log_error_write (srv, __FILE__, __LINE__, "s",
				"FREE_FUNC: plugin_data has no data");
		return HANDLER_GO_ON;
	}

	if ( p->config_storage ) {
		size_t i;
		for ( i = 0; i < srv->config_context->used; i++ ) {
			plugin_config *s = p->config_storage[i];

			if ( ! s ) continue;

			buffer_free (s->order);
			array_free (s->list);

			free (s);
		}
		free (p->config_storage);
	}

	free (p);
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC (mod_net_access_set_defaults) {
	plugin_data *p = p_d;
	size_t i;

	config_values_t cv[] = {
		{ "naccess.debug", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },
		{ "url.order", NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },
		{ "url.list",  NULL, T_CONFIG_ARRAY, T_CONFIG_SCOPE_CONNECTION },
		{ NULL,            NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
	};

	if ( ! p ) {
		log_error_write (srv, __FILE__, __LINE__, "s", "can't initionalize plugin_data");
		return HANDLER_ERROR;
	}

	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));

	for ( i=0; i < srv->config_context->used; i++ ) {
		plugin_config *s;

		s = calloc (1, sizeof (plugin_config));
		s->debug = 0;
		s->order = buffer_init ();
		s->list = array_init ();

		cv[0].destination = &(s->debug);
		cv[1].destination = s->order;
		cv[2].destination = s->list;

		p->config_storage[i] = s;

		if ( 0 != config_insert_values_global (srv, ((data_config *) srv->config_context->data[i])->value, cv) ) {
			log_error_write (srv, __FILE__, __LINE__, "s", "Can't insert global config value");
			return HANDLER_ERROR;
		}
	}

	return HANDLER_GO_ON;
}

#define PATCH(x) \
	p->conf.x = s->x;
static int mod_net_access_patch_connection (server *srv, connection *con, plugin_data *p) {
	plugin_config *s = p->config_storage[0];
	size_t i, j;

	PATCH (debug);
	PATCH (order);
	PATCH (list);

	/* skip the first, the global context */
	for ( i=1; i < srv->config_context->used; i++ ) {
		data_config *dc = (data_config *) srv->config_context->data[i];
		s = p->config_storage[i];

		/* condition didn't match */
		if ( ! config_check_cond (srv, con, dc) ) continue;

		/* merge config */
		for ( j=0; j<dc->value->used; j++ ) {
			data_unset *du = dc->value->data[j];

			if ( buffer_is_equal_string (du->key, CONST_STR_LEN("naccess.debug")) ) {
				PATCH (debug);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("url.order")) ) {
				PATCH (order);
			} else if ( buffer_is_equal_string (du->key, CONST_STR_LEN("url.list")) ) {
				PATCH (list);
			}

		}
	}

	if ( p->conf.debug ) {
		log_error_write (srv, __FILE__, __LINE__, "ss", "url.order:", p->conf.order->ptr);
		for ( i=0; i<p->conf.list->used; i++ ) {
			data_string *ds = (data_string *) p->conf.list->data[i];
			log_error_write (srv, __FILE__, __LINE__, "ss", "url.list:", ds->value->ptr);
		}
	}

	return 0;
}
#undef PATCH

URIHANDLER_FUNC (mod_net_access_uri_handler) {
	plugin_data *p = p_d;
	unsigned short handler_naccess;

	/* select the right config */
	mod_net_access_patch_connection (srv, con, p);

	if ( p->conf.order->used == 0 || ! strcmp (p->conf.order->ptr, "none") )
		return HANDLER_GO_ON;

	if ( ! naccess_match (srv, con, p->conf) ) {
		/* SUCCESS Match */
		if ( ! strcmp (p->conf.order->ptr, "allow") )
			handler_naccess = HANDLER_GO_ON;
		else
			handler_naccess = HANDLER_FINISHED;
	} else {
		/* Failed Match */
		if ( ! strcmp (p->conf.order->ptr, "deny") )
			handler_naccess = HANDLER_GO_ON;
		else
			handler_naccess = HANDLER_FINISHED;
	}

	if ( handler_naccess == HANDLER_FINISHED ) {
		con->http_status = 403;
		con->mode = DIRECT;
	}

	return handler_naccess;
}

/* this function is called at dlopen() time and inits the callbacks */
int mod_net_access_plugin_init (plugin *p) {
	p->version          = LIGHTTPD_VERSION_ID;
	p->name             = buffer_init_string ("net_access");

	p->init             = mod_net_access_init;
	p->set_defaults     = mod_net_access_set_defaults;
	p->handle_uri_clean = mod_net_access_uri_handler;

	p->cleanup          = mod_net_access_free;

	p->data             = NULL;

	return 0;
}

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
